'''

will use https://github.com/stanstrup/commonMZ for now, which come mostly from Keller et al, 2008.

Could consider the tables from CAMERA later.

Two sets of tables for pos and neg ESI.

'''
import os
from collections import namedtuple
from itertools import combinations

import numpy as np
import matplotlib.pyplot as plt
from sklearn.neighbors import KernelDensity

from pyopenms import MSExperiment, MzMLFile

# Define major data strctures

Scan = namedtuple('Scan', [
    'scan_number', 'retention_time', 'mz_values', 'intensity_values'
    ])

class Sample: # modified from asari
    '''
    A sample or injection, corresponding to one raw data file, either from positive or negative ionization.
    '''
    def __init__(self, experiment, mode, input_file=''):
        self.input_file = input_file
        self.experiment = experiment                        # parent Experiment instance
        self.name = os.path.basename(input_file)            # must be unique
        self.mode = mode            # ESI mode, could switch within expt
        
        # list of scans, each as one spectrum collected at a specific elution time
        self.scans = []
        
        # these two lists are in mached order; will make immutable tuples after reading data
        self.retention_index = []     # this corresponds to scan number
        self.retention_time = []
        


def get_pairwise_diffs(L):
    # limit to diff btw (1,100) for efficiency. E.g. scan #1 234955 -> 
    d = [abs(x-y) for x,y in combinations(L,2)]
    return [x for x in d if 1<x<100]

def read_sample(input_file, expt_='ecoli', mode='pos'): 
    SM = Sample(expt_, mode, input_file) 
    exp = MSExperiment()
    MzMLFile().load(input_file, exp)
    print(input_file, exp.getNrSpectra())
    scans = []
    ii = 0
    for sp in exp:
        ms_level = sp.getMSLevel()
        if ms_level == 1:    # ONLY dealig with MS1 here
            rt = sp.getRT()
            mz, intensity = sp.get_peaks()
            # use int and tuple to save space when storage is considered
            intensity = [int(x) for x in intensity]
            # Scan defined as 'scan_number', 'retention_time', 'mz_values', 'intensity_values'
            scans.append(Scan(ii, rt, tuple(mz), tuple(intensity)))
            # scans.append(Scan(ii, rt, mz, intensity))
        ii += 1
        
    SM.scans = scans
    SM.retention_index = tuple([sc.scan_number for sc in scans])
    SM.retention_time = tuple([sc.retention_time for sc in scans])
    return SM











'''
DB based, step wise, formula mass centered MS1 annotation

use case 1:
    input feature table
    output:
        feature table with flags of contaminants, empCpd and corresponding ions
        nonredundant table with one feature (selected by highest intensity) per empCpd, and empCpd annotations

use case 2:
    single sample annotation, to faciliate correspondence in data preprocessing


To-do:
# to import library of authentic chemical standards

# core data structures
# from metDataModel import derived

Old data structure in mummichog v2:
"C05769": {"formula": "C36H34N4O8", "mw": 654.269, "name": "Coproporphyrin I", 
"adducts": {"M+2H[2+]": 328.14177646677, "M+Br81[-]": 735.1853, "M-H2O+H[1+]": 637.2656764667701, "M-C3H4O2+H[1+]": 583.25517646677, "M-HCOOH+H[1+]": 609.27087646677, "M-CO+H[1+]": 627.28127646677, 
        "M+K[1+]": 693.23177646677, "M+Cl[-]": 689.2379, "M+Na-2H[-]": 674.23644706646, "M-CO2+H[1+]": 611.28647646677, "M+Na[1+]": 677.25827646677, "M-2H[2-]": 326.12722353323, "M+H[1+]": 655.27627646677, 
        "M-H4O2+H[1+]": 619.25507646677, "M(C13)-H[-]": 654.26512353323, "M+HCOONa[1+]": 723.26367646677, "M(C13)+2H[2+]": 328.64347646677004, "M+HCOOK[1+]": 739.23757646677, "M+HCOO[-]": 699.266645, 
        "M(C13)+3H[3+]": 219.43134313343666, "M-H[-]": 653.26172353323, "M+ACN-H[-]": 694.2882685332299, "M+Cl37[-]": 691.2349, "M-H2O-H[-]": 635.25112353323, "M+Br[-]": 733.1873, 
        "M+3H[3+]": 219.09694313343667, "M+CH3COO[-]": 713.282295, "M(C13)+H[1+]": 656.2796764667701, "M[1+]": 654.269, "M-NH3+H[1+]": 638.2497764667701, "M+NaCl[1+]": 713.2348764667701, 
        "M+H+Na[2+]": 339.13277646677, "M+H2O+H[1+]": 673.28687646677, "M-H+O[-]": 669.25663353323, "M+K-2H[-]": 690.20994706646}}, 

Use DB1, DB2, (DB3, ...) for tiered searches.
Start with DB1 for searching primary ions, because other adducts/isotopes can only exist if a primary ion does.
Precompute selectivity within each DB. Use in indexed format.

Use indexed_DB            
example DB_1[99] = 
            [['C2H4NaO3_99.00532', 99.00532046676999, 'C2H4NaO3', 0.9999999571942034, [('C2H4O3_76.016044', 'M+Na[1+]')]], 
            ['C4H5NS_99.013721', 99.01372099999999, 'C4H5NS', 0.9967247109333564, [('C4H5NS_99.01427', 'M[1+]')]], 
            ['C3H8NaO2_99.041706', 99.04170646677, 'C3H8NaO2', 0.999999999998776, [('C3H8O2_76.05243', 'M+Na[1+]')]], ...]

empCpds will be organized based on matched DB entries after formula_mass annotation.


'''

# import re
import numpy as np
# from scipy.optimize import curve_fit
from scipy.stats import norm as normal_distribution

# from metDataModel.core import import Feature, EmpiricalCompound

# databases
from .lib.common_mass import mass_signatures
from .lib.massDicts import massDict_hmdb_serum, massDict_hmdb
from .lib.PubChemLite import PubChemLite

from .lib.LCMS_contaminants import contaminants_pos, contaminants_neg

from .io import tFeature, read_feature_tuples
from .formula import compute_adducts_formulae

# from .calibrate import mass_calibrate


# ad hoc, will repalce with better lists for calibration, for each of pos and neg.
calibration_mass_dict_pos = {'C4H9N3O2_131.069477': 132.07675346677001, 'C6H11NO2_129.078979': 130.08625546677, 'C5H9NO4_147.053158': 148.06043446677, 'C6H10O6_178.047738': 179.05501446677002, 'C9H11NO3_181.073893': 182.08116946677, 'C9H11NO2_165.078979': 166.08625546677, 'C9H8O3_164.047344': 165.05462046677002, 'C23H45NO4_399.334859': 400.34213546677, 'C2H7NO3S_125.014664': 126.02194046676999, 'C5H7NO3_129.042593': 130.04986946677002, 'C5H4N4O3_168.02834': 169.03561646677, 'C5H8O5_148.037173': 149.04444946677, 'C5H12O5_152.068473': 153.07574946677002, 'C9H8O2_148.052429': 149.05970546677, 'C5H10N2O3_146.069142': 147.07641846677, 'C17H33NO4_315.240959': 316.24823546677, 'C5H11NO2S_149.051049': 150.05832546677001, 'C5H10O2_102.06808': 103.07535646676999, 'C15H29NO4_287.209658': 288.21693446677, 'C8H17NO2_159.125929': 160.13320546677002, 'H2O4S_97.967379': 98.97465546676999, 'C6H13NO5_179.079373': 180.08664946677, 'C16H22O4_278.151809': 279.15908546677, 'C20H28O2_300.20893': 301.21620646677, 'C7H8N4O2_180.064726': 181.07200246677002, 'C7H6O2_122.036779': 123.04405546676999, 'C14H22N2O3_266.163043': 267.17031946677, 'C19H16O4_308.104859': 309.11213546677, 'C21H39NO4_369.287909': 370.29518546677, 'C8H6O4_166.026609': 167.03388546677002, 'C18H35NO_281.271865': 282.27914146677, 'C19H37NO4_343.272259': 344.27953546677, 'C26H52NO7P_521.34814': 522.35541646677, 'C7H8N2O2_152.058578': 153.06585446677002, 'C9H17NO2_171.125929': 172.13320546677002, 'C25H47NO4_425.350509': 426.35778546677, 'C8H10O3_154.062994': 155.07027046677, 'C25H45NO4_423.334859': 424.34213546677, 'C22H46NO7P_467.301189': 468.30846546677003, 'C23H48NO7P_481.316839': 482.32411546677, 'C24H48NO7P_493.316839': 494.32411546677, 'C26H54NO7P_523.36379': 524.37106646677, 'C26H48NO7P_517.316839': 518.32411546677, 'C28H52NO7P_545.34814': 546.35541646677, 'C28H50NO7P_543.332489': 544.33976546677, 'C24H50NO6P_479.337575': 480.34485146677, 'C25H52NO7P_509.34814': 510.35541646677, 'C21H38O4_354.27701': 355.28428646677, 'C17H31NO4_313.225308': 314.23258446677, 'C15H27NO4_285.194008': 286.20128446677, 'C19H35NO4_341.256609': 342.26388546677003, 'C22H37NO3_363.277344': 364.28462046677004, 'C19H22ClN5O_371.151288': 372.15856446677003, 'C24H30O6_414.204239': 415.21151546677, 'C21H25N_291.1987': 292.20597646677, 'C15H23NO2_249.172879': 250.18015546677, 
    'C17H27NO4_309.194008': 310.20128446677, 'C21H25ClN2O3_388.15537': 389.16264646677, 'C16H18O10_370.089997': 371.09727346677}


MASS_RANGE = (50, 2000)
# fraction of total retention time, or of ranks of retention time
# used to determine coelution of ions ad hoc
RETENTION_TIME_TOLERANCE_FRAC = 0.02

FEATURE_REGISTRY = {}
EMPCPD_REGISTRY = {}


def make_formula_mass_id(formula, mass):
    return formula + '_' + str(round(mass,6))

def list_search_formula_mass_db(list_query_mz, indexed_DB, limit_ppm=10):
    return [search_formula_mass_db(query_mz, indexed_DB, limit_ppm) for query_mz in list_query_mz]

def bin_by_median(List_of_tuples, func_tolerance):
    '''
    Not perfect because left side may deviate out of tolerance, but LC-MS data always have enough gaps for separation.
    Will add kernel density method for grouping m/z features.
    List_of_tuples: [(value, object), (value, object), ...], to be separated into bins by values (either rt or mz).
                    objects have attribute of sample_name if to align elsewhere.
    return: [seprated bins], each as a list in the same input format. If all falls in same bin, i.e. [List_of_tuples].
    '''
    new = [[List_of_tuples[0], ], ]
    for X in List_of_tuples[1:]:
        if X[0]-np.median([ii[0] for ii in new[-1]]) < func_tolerance(X[0]):       # median moving with list change
            new[-1].append(X)
        else:
            new.append([X])
    PL = []
    for L in new:
        PL.append([X[1] for X in L])
    return PL


def assemble_empcpds_by_formula(List_of_features):
    pass


def assemble_empcpds_denovo(List_of_features):
    pass


def search_formula_mass_db(query_mz, indexed_DB, limit_ppm=10):
    '''
    Find best matched formula_mass in indexed_DB for query_mz within ppm limit.
    
    indexed_DB: # [formula_mass, m/z, charged_formula, selectivity, [relations to neutral formula]]
            example DB_1[99] = 
            [['C2H4NaO3_99.00532', 99.00532046676999, 'C2H4NaO3', 0.9999999571942034, [('C2H4O3_76.016044', 'M+Na[1+]')]], 
            ['C4H5NS_99.013721', 99.01372099999999, 'C4H5NS', 0.9967247109333564, [('C4H5NS_99.01427', 'M[1+]')]], 
            ['C3H8NaO2_99.041706', 99.04170646677, 'C3H8NaO2', 0.999999999998776, [('C3H8O2_76.05243', 'M+Na[1+]')]], ...]

    return
    ------
    Closest match as an entry,
             (match_ppm, [formula_mass, m/z, charged_formula, selectivity, [relations to neutral formula]])
    None if out of ppm limit.
    '''
    _delta = query_mz * limit_ppm * 0.000001
    _low_lim, _high_lim = query_mz - _delta, query_mz + _delta
    result = []
    for ii in range(int(_low_lim), int(_high_lim)+1):
        if ii in indexed_DB:
            for F in indexed_DB[ii]:
                # F[1] is m/z, fixed DB format
                result.append( (abs(query_mz-F[1]), F) )

    result.sort()
    if result and result[0][0] < _delta:
        return result[0][1]
    else:
        return None

def create_extended_DB(list_empCpd_seeds, mode='pos'):
    '''
    This creates a new DB of common isotopes and adducts based on a given seed list.
    This is using ions specified in .formula.compute_adducts_formulae.
    This can be used on what's found after getting primary ions in DB1, 
    because isotopes/adducts depend on the existence of a primary ion.
    Otherwise, the presence of a primary ion will be enforced at empCpd level.

    input
    -----
    list_empCpd_seeds: list of entries as [formula_mass, ...]. Not full empCpd class.
    
    return
    ------
    indexed_DB in the format of each entry like:
    [formula_mass, m/z, charged_formula, selectivity, [relations to neutral formula]]

    '''
    new_DB = {}
    for x in set(list_empCpd_seeds):             # remove redundancy
        neutral_formula, mw = x.split("_")       # [('C2H4O3_76.016044',..
        adducts = compute_adducts_formulae(float(mw), neutral_formula, mode)    
                                                 # e.g. [(58.53894096677, 'M+2H[2+]', result_formula), ...,]
        for A in adducts:
            new = [ make_formula_mass_id(A[2], A[0]), A[0], A[2], None, [(x, A[1]),] ]
            ii = int(A[0])
            if ii in new_DB:
                new_DB[ii].append(new)
            else:
                new_DB[ii] = [new]

    return new_DB


def annotate_formula_mass(list_query_mz, indexed_DB, check_mass_accuracy=False, limit_ppm=10):
    '''
    Annotate input m/z list using formula based mass in the indexed_DB,
    which should be generic, not assuming separation of primary ions and their common isotopes and adducts.
    Search efficiency will be improved by the construction strategy of the indexed_DB,
    not dependent on the algorithm in this function. E.g., more frequent features should be in first DB to search.

    If check_mass_accuracy is True, the ppm deviations of each matched primary ion are fitted to a normal distribution,
    to estimate mass_accuracy and ppm_std.

    input
    -----
    input_list: m/z values.
    indexed_DB: per entry - [formula_mass, m/z, charged_formula, ion, selectivity, [relations to neutral formula]]
                This can be a pre-loaded DB, or hot DB in asari.

    # mode='pos', 

    return
    ------
    {mass_accuracy: estimated based on difference from formula value
    ppm_std:         estimated based on precision distribution
    annotated_list: [matched DB entry or None, ...]
    corrected_list_query_mz: m/z values will be corrected if they deviate from formula values by mass_accuracy > 5 ppm.
    }
    '''
    result = list_search_formula_mass_db(list_query_mz, indexed_DB, limit_ppm)
    mass_accuracy, ppm_std, corrected_list_query_mz = None, None, None
    N = len(list_query_mz)
    if check_mass_accuracy:
        list_ppm_errors = []
        for ii in range(N):
            if result[ii]:
                list_ppm_errors.append( 1000000 * (list_query_mz[ii] - result[ii][1])/result[ii][1] )

        mass_accuracy, ppm_std = normal_distribution.fit(list_ppm_errors)
        if abs(mass_accuracy) > 5:   # this is considered significant mass shift, requiring m/z correction for all 
            list_query_mz = [x-x*0.000001*mass_accuracy for x in list_query_mz]
            # redo search because we may find different matches after mass correction
            # Also change search ppm if self.__mass_stdev__ > 5
            result = list_search_formula_mass_db(list_query_mz, indexed_DB, max(10, 2*ppm_std))
            corrected_list_query_mz = list_query_mz
            # update ppm_std because it may be used for downstream search parameters
            list_ppm_errors = []
            for ii in range(N):
                if result[ii]:
                    list_ppm_errors.append( 1000000 * (list_query_mz[ii] - result[ii][1])/result[ii][1] )
            _mu, ppm_std= normal_distribution.fit(list_ppm_errors)

        elif ppm_std > 5:      # no mass correction but need redo search using higher ppm for unmatched mz
            result = list_search_formula_mass_db(list_query_mz, indexed_DB, 2*ppm_std)

    return {
        'mass_accuracy': mass_accuracy,
        'ppm_std': ppm_std,
        'annotated_list': result,
        'corrected_list_query_mz': corrected_list_query_mz,
        }


# -----------------------------------------------------------------------------
#
# Prototype code with short-handed data structure, 
# still used for particular cases, backward compatibility only during dev
#
# -----------------------------------------------------------------------------

def do_two_step_annotation(indexed_features, anchorDict, mode='pos',  
                                            MASS_RANGE=MASS_RANGE, ppm=2):
    '''
    Wrapper of annotate_anchorList and extend_annotate_anchorList.
    '''
    _empCpds, _unmatched_features = annotate_anchorList(
                                            indexed_features, anchorDict, ppm)
    _empCpds, _unmatched_features = extend_annotate_anchorList( 
                                            _empCpds, _unmatched_features, mode,  MASS_RANGE, ppm)
    print("Got matched empCpds: ", len(_empCpds))

    return _empCpds, _unmatched_features


def annotate_anchorList(indexed_features, anchorList, ppm=2):
    '''
    Annotate indexed_features by an anchor set of formula_mass values.
    
    Input
    -----
    indexed_features: {int_mz: [{'id': '', 'mz': 0, 'calibrated_mz': 0, 'rtime': 0, ...}, ...],
                                ...}
    anchorList: {'C7H11N3O2_169.085127': [(170.09240346677, '+H[1+]'), (192.074345, 'Na'), (169.085127, 'M')], ...}
    ppm: Based on m/z only, but for calibrated m/z, this should be very stringent, thus ppm default at 2.

    Return
    -------
    matched_empCpds: List of empCpds, e.g. below.
    unmatched_features: Features not matched to anchorList, unindexed
    
    Example
    -------
    matched_empCpds, unmatched_features = annotate_anchorList(my_cal_flaged, anchorList, 2)

    726 702 3939

    matched_empCpds[33]

    {'list_of_features': [{'id': '"FT0394"',
    'mz': 116.070646922025,
    'rtime': 36.3316230773926,
    'intensities': [],
    'calibrated_mz': 116.07051083925319,
    'potential_contaminants': [],
    'ion': '+H[1+]'},
    {'id': '"FT0670"',
    'mz': 138.052651169964,
    'rtime': 36.1285610198975,
    'intensities': [],
    'calibrated_mz': 138.052503250255,
    'potential_contaminants': [],
    'ion': 'Na'},
    {'id': '"FT0379"',
    'mz': 115.063235073422,
    'rtime': 31.9573593139648,
    'intensities': [],
    'calibrated_mz': 115.06309953312442,
    'potential_contaminants': [],
    'ion': 'M'}],
    'interim_id': 'C5H9NO2_115.063329',
    'neutral_formula': 'C5H9NO2',
    'neutral_base_mass': 115.063329}

    len(matched_empCpds), len(unmatched_features)

    (504, 3939)

    # this example shows 504 empCpds are found in 4641 features.
    # The empCpds include 702 Features, of which 24 have multiple matches.
    '''

    matched_empCpds, unmatched_features, matched_features = [], [], []
    for anchor_compound in list(anchorList.items()):
        _m = __anchor_to_empCpd__(indexed_features, anchor_compound, ppm)
        if _m:
            matched_empCpds.append(_m)

    for F in matched_empCpds: 
        matched_features += [x['id'] for x in F['list_of_features']]

    for k,v in indexed_features.items():
        for F in v:
            if F['id'] not in matched_features:
                unmatched_features.append(F)

    print("Remaining unmatched featuress: ", len(unmatched_features))
    return matched_empCpds, unmatched_features


def extend_annotate_anchorList(list_empCpds, unmatched_features, mode='pos',  MASS_RANGE=MASS_RANGE, ppm=2):
    '''
    Annotate unmatched_features by extended adducts.
    
    Input
    -----
    list_empCpds: List of anchor matched empCpds
    unmatched_features: [{'id': '', 'mz': 0, 'calibrated_mz': 0, 'rtime': 0, ...}, ...],
    mode: ionization mode, pos or neg.
    ppm: Based on m/z only, but for calibrated m/z, this should be very stringent, thus ppm default at 2.

    Return
    -------
    matched_empCpds: List of updated empCpds, 
    updated_unmatched_features: Features not in matched_empCpds, unindexed
    '''

    unmatched_features = index_features(unmatched_features)
    return __extend_search__(list_empCpds, unmatched_features, mode,  MASS_RANGE, ppm)





def flag_contaminants(indexed_features, contaminant_list, ppm=2):
    '''
    Flag features if they match to known contaminants;
    set F['potential_contaminants'] = [contaminants, ].
    Based on m/z only, but for calibrated m/z, this should be very stringent, thus ppm default at 2.

    contaminant_list is dependent on ion mode, pos or neg. There's some redundancy in the lists.
    # mass, good_name, name, formula, ion_form, possible origin
    In [14]: contaminants_pos[:2]                                                                                             
    Out[14]: 
    [['537.8790134',
    'C2H4O2_[M6-H6+Fe3+O]+_537.879013',
    'Acetic Acid',
    'C2H4O2',
    '[M6-H6+Fe3+O]+',
    'Solvent'],
    ['555.8895784',
    'C2H4O2_[M6-H6+H2O+Fe3+O]+_555.889578',
    'Acetic Acid',
    'C2H4O2',
    '[M6-H6+H2O+Fe3+O]+',
    'Solvent']]

    This can be faster, as now looping unindexed contaminant_list

    Input
    -----
    indexed_features: {int_mz: [{'id': '', 'mz': 0, 'calibrated_mz': 0, 'rtime': 0, ...}, ...],
                                ...}

    '''
    def check_contaminant(mz, contaminant_list, ppm):
        matched = []
        for T in contaminant_list:
            if abs(T[0] - mz)/mz < 0.000001*ppm:
                matched.append(T)
        return matched

    for k,v in indexed_features.items():
        for F in v:
            F['potential_contaminants'] = check_contaminant(F['calibrated_mz'], contaminant_list, ppm)

    return indexed_features


def make_anchordict(massDict_db, mode='pos', MASS_RANGE=MASS_RANGE):
    '''
    To build anchor mz list based on keys like 'C7H11N3O2_169.085127'.
    This core list is used to separate ions from a feature list to reduce search space.
    The matched results will be organized by the same key of formula_mass.

    The core list is designed for mandating the presence of one of the primary ions, 
    without which other adducts won't be considered.

    Input
    -----
    massDict_db: using formula_mass as identifiers, e.g. 
                'C7H11N3O2_169.085127':[['HMDB0000001', '1-Methylhistidine'], ['HMDB0000479', '3-Methylhistidine']]

    Return
    ------
    {'C7H11N3O2_169.085127': [(170.09240346677, '+H[1+]'), (192.074345, 'Na'), (169.085127, 'M')], ...}
    '''
    posList = [(1.00727646677, 'M+H[1+]'), (22.989218, 'Na'), (-0.0005, 'M*'), 
                (19.01787646677, 'M+H2O+H[1+]')]
    # ? 'M-2H[2-]'
    negList = [(-1.00727646677, 'M-H[-]'), (34.969402, 'M+Cl-'), (0.0005, 'M+e[-]'), 
                (-19.01787646677, 'M-H2O-H[-]')]

    if mode == 'pos': useList = posList
    elif mode == 'neg': useList = negList
    else: print ("ionization mode is either pos or neg.")

    new = {}
    for k in massDict_db:
        kmass = float(k.split('_')[1])
        new[k] = [(x+kmass, y) for x,y in useList if MASS_RANGE[0] < x+kmass < MASS_RANGE[1]]
    return new

def search_feature_pair(F1, F2, mass_signatures, ppm=2):
    '''
    # unidirectional, F2-F1 only
    # msg is like (1.00727646677, '+H[1+]'),

    return
    ------
    list of matched: [('"FT1104"', '"FT1630"', 43.96389, '2Na-2H'), ...]
    '''
    _matched = []
    for msg in mass_signatures:
        if abs( F2['calibrated_mz'] - F1['calibrated_mz'] - msg[0] ) / F2['calibrated_mz'] < 0.000001*ppm:
            _matched.append( (F1['id'], F2['id'], msg[0], msg[1]) )
            
    return _matched


# -----------------------------------------------------------------------------
#
# Functions below are for internal use. Don't touch.
#
# -----------------------------------------------------------------------------


def __find_mz_indexed_features__(query_mz, indexed_features, limit_ppm=2):
    # to find all matches of a m/z in indexed_features, under limit_ppm
    matched = []
    _delta = query_mz * limit_ppm * 0.000001
    low_lim, high_lim = query_mz - _delta, query_mz + _delta
    for ii in range(int(low_lim), int(high_lim)+1):
        for F in indexed_features[ii]:
            if low_lim < F['calibrated_mz'] < high_lim:
                matched.append(F)

    return matched

def __anchor_to_empCpd__(indexed_features, anchor_compound, ppm):
    '''
    Annotate indexed_features into empCpd.
    EmpiricalCompound = {
        "neutral_base_mass": 0,
        "neutral_formula": [],
        "interim_id": '', 
        "list_of_features": [
                {'id': '', 'ion': 'M-H[1-]', 'm/z': 169.0013, 'rtime': 55, ..},
                {},
    }
    
    Input
    -----
    indexed_features: {int_mz: [{'id': '', 'mz': 0, 'calibrated_mz': 0, 'rtime': 0, ...}, ...],
                                ...}
    anchor_compound: ('C7H11N3O2_169.085127', [(170.09240346677, '+H[1+]'), (192.074345, 'Na'), (169.085127, 'M')])
    ppm: Based on m/z only, but for calibrated m/z, this should be very stringent, thus ppm default at 2.

    Return
    -------
    empCpd if matched; otherwise {}
    ''' 
    _matched = False
    tmp = {
        "list_of_features": [],
    }
    for v in anchor_compound[1]:
        found_features = __find_mz_indexed_features__(v[0], indexed_features, ppm)
        if found_features:
            _matched= True
            for F in found_features: F['ion'] = v[1]
            tmp['list_of_features'] += found_features

    if _matched:
        tmp['interim_id'] = anchor_compound[0]
        neutral_formula, neutral_base_mass = anchor_compound[0].split('_')
        tmp['neutral_formula'], tmp['neutral_base_mass'] = neutral_formula, float(neutral_base_mass)
        return tmp
    else:
        return None



def __extend_search__(list_empCpds, indexed_unmatched_features, 
                            mode='pos',  MASS_RANGE=MASS_RANGE, ppm=2):
    '''
    Annotate indexed_features by extended adducts.
    
    Input
    -----
    list_empCpds: List of anchor matched empCpds
    indexed_unmatched_features: indexed_features as {int_mz: [{'id': '', 'mz': 0, 'calibrated_mz': 0, 'rtime': 0, ...}, ...],
                                ...}
    mode: ionization mode, pos or neg.
    ppm: Based on m/z only, but for calibrated m/z, this should be very stringent, thus ppm default at 2.

    Return
    -------
    matched_empCpds: List of updated empCpds, 
    updated_unmatched_features: Features not in matched_empCpds, unindexed

    '''
    _tmp_matched = []
    for E in list_empCpds:
        for target_mz, note, result_formula in compute_adducts_formulae(E['neutral_base_mass'], E['neutral_formula'], mode) :
            if MASS_RANGE[0] < target_mz < MASS_RANGE[1]:
                found_features = __find_mz_indexed_features__(target_mz, indexed_unmatched_features, ppm)
                for F in found_features: 
                    _tmp_matched.append(F['id'])
                    # format?
                    F['ion'] = (result_formula, note)
                    E['list_of_features'].append(F)

    updated_unmatched_features = []
    for k,v in indexed_unmatched_features.items():
        for F in v:
            if F['id'] not in _tmp_matched:
                updated_unmatched_features.append(F)

    print("Remaining unmatched featuress: ", len(updated_unmatched_features))
    return list_empCpds, updated_unmatched_features







#
# -----------------------------------------------------------------------------
#
# not used now
# 

def find_isotopic_pairs(list_peaks, mztree, isotopic_pair, mz_tolerance_ppm=5, rt_tolerance_scans=5):
    '''
    - to update

    To find naturally occuring isotopic pairs in LC-MS peaks, 
    using strict coelution criteria and loose abundance ratio.
    This is done within a sample, so that all elution time is precisely matched by scan numbers.
    This avoids any issues in alignment across samples.
    Rules: elution apex should be within 5 scans; overlap of elution peaks minimal 50% of the smaller peak;
            abundance ratio loosely applied, leaving room for measuring errors.
    
    Abundance of natually occuring isotopic elements:
    12C~99%, 13C ~ 1%
    14N ~ 99.64%, 15N ~ 0.36%
    Sulfur:  32S (95.02%), 33S (0.75%), 34S (4.21%), 
    Chlorine has two stable isotopes, 35Cl (75.77%) and 37Cl (24.23%).
    1H and 31P are basically ~100% in nature.

    Input
    =====
    list_peaks: [{'parent_masstrace_id': 1670, 'mz': 133.09702315984987, 'apex': 654, 'height': 14388.0, 
    'left_base': 648, 'right_base': 655, 'id_number': 555},
                 ...]
    mztree is constructed on list_peaks.
    isotopic_pair: (mass_difference, notion, abundance_ratio). E.g. (1.0034, 'M(C13)', 0.2). - to update
    
    Return
    ======
    list of pairs of peaks.
    '''
    pairs = []
    (mass_difference, relation, abundance_ratio_min, abundance_ratio_max) = isotopic_pair
    for P1 in list_peaks:
        tmp = find_all_matches_centurion_indexed_list(P1['mz'] + mass_difference, mztree, mz_tolerance_ppm)
        for P2 in tmp:
            if abs(P1['apex']-P2['apex']) <= rt_tolerance_scans and is_coeluted(P1, P2) \
                                             and abundance_ratio_min*P1['height'] < P2['height'] < abundance_ratio_max*P1['height']:
                pairs.append((P1, P2))

    return pairs


def init_eTrees_by_isotopic_signatures(isotopic_signatures, iso_type, peak_list):
    '''
    An example of isotopic_signatures: 
    [(296, 'anchor'), (339, 'M(13C)'), (332, 'M(15N)'), (333, 'M(15N)')], using peak numbers. 
    

    '''
    eTrees = []
    (P1, P2) = iso_pairs
    for pp in iso_pairs:
        ET = epdTree(peak_list)
        ET.anchor_peak = P1['id_number']
        ET.isotope_peaks = [P1['id_number'], P2['id_number']]
        ET.peak_relationships.append(( P1['id_number'], P2['id_number'], iso_type ))
    return eTrees

def merge_isotopic_eTrees(list_eTrees):
    '''
    For each pair of eTrees after init_eTrees_by_isopairs, merge if at least one overlap peak is anchor. 
    Ignor if overlap is all at endpoints. I.e. trees can branch, but not converge.
    '''
    new = []
    peak2etree = {}
    for ET in list_eTrees:
        for P in ET.isotope_peaks:
            if P in peak2etree:
                peak2etree[P].append()

    pass



#
# -----------------------------------------------------------------------------
#



def search_formula_mass_db(query_mz, indexed_DB, limit_ppm=10):
    '''
    Find best matched formula_mass in indexed_DB for query_mz within ppm limit.
    
    indexed_DB: # [formula_mass, m/z, charged_formula, selectivity, [relations to neutral formula]]
            example DB_1[99] = 
            [['C2H4NaO3_99.00532', 99.00532046676999, 'C2H4NaO3', 0.9999999571942034, [('C2H4O3_76.016044', 'M+Na[1+]')]], 
            ['C4H5NS_99.013721', 99.01372099999999, 'C4H5NS', 0.9967247109333564, [('C4H5NS_99.01427', 'M[1+]')]], 
            ['C3H8NaO2_99.041706', 99.04170646677, 'C3H8NaO2', 0.999999999998776, [('C3H8O2_76.05243', 'M+Na[1+]')]], ...]

    return
    ------
    Closest match as an entry,
             (match_ppm, [formula_mass, m/z, charged_formula, selectivity, [relations to neutral formula]])
    None if out of ppm limit.
    '''
    _delta = query_mz * limit_ppm * 0.000001
    _low_lim, _high_lim = query_mz - _delta, query_mz + _delta
    result = []
    for ii in range(int(_low_lim), int(_high_lim)+1):
        if ii in indexed_DB:
            for F in indexed_DB[ii]:
                # F[1] is m/z, fixed DB format
                result.append( (abs(query_mz-F[1]), F) )

    result.sort()
    if result and result[0][0] < _delta:
        return result[0][1]
    else:
        return None



def annotate_formula_mass(list_query_mz, indexed_DB, check_mass_accuracy=False, limit_ppm=10):
    '''
    Annotate input m/z list using formula based mass in the indexed_DB,
    which should be generic, not assuming separation of primary ions and their common isotopes and adducts.
    Search efficiency will be improved by the construction strategy of the indexed_DB,
    not dependent on the algorithm in this function. E.g., more frequent features should be in first DB to search.

    If check_mass_accuracy is True, the ppm deviations of each matched primary ion are fitted to a normal distribution,
    to estimate mass_accuracy and ppm_std.

    input
    -----
    input_list: m/z values.
    indexed_DB: per entry - [formula_mass, m/z, charged_formula, ion, selectivity, [relations to neutral formula]]
                This can be a pre-loaded DB, or hot DB in asari.

    # mode='pos', 

    return
    ------
    {mass_accuracy: estimated based on difference from formula value
    ppm_std:         estimated based on precision distribution
    annotated_list: [matched DB entry or None, ...]
    corrected_list_query_mz: m/z values will be corrected if they deviate from formula values by mass_accuracy > 5 ppm.
    }
    '''
    result = list_search_formula_mass_db(list_query_mz, indexed_DB, limit_ppm)
    mass_accuracy, ppm_std, corrected_list_query_mz = None, None, None
    N = len(list_query_mz)
    if check_mass_accuracy:
        list_ppm_errors = []
        for ii in range(N):
            if result[ii]:
                list_ppm_errors.append( 1000000 * (list_query_mz[ii] - result[ii][1])/result[ii][1] )

        mass_accuracy, ppm_std = normal_distribution.fit(list_ppm_errors)
        if abs(mass_accuracy) > 5:   # this is considered significant mass shift, requiring m/z correction for all 
            list_query_mz = [x-x*0.000001*mass_accuracy for x in list_query_mz]
            # redo search because we may find different matches after mass correction
            # Also change search ppm if self.__mass_stdev__ > 5
            result = list_search_formula_mass_db(list_query_mz, indexed_DB, max(10, 2*ppm_std))
            corrected_list_query_mz = list_query_mz
            # update ppm_std because it may be used for downstream search parameters
            list_ppm_errors = []
            for ii in range(N):
                if result[ii]:
                    list_ppm_errors.append( 1000000 * (list_query_mz[ii] - result[ii][1])/result[ii][1] )
            _mu, ppm_std= normal_distribution.fit(list_ppm_errors)

        elif ppm_std > 5:      # no mass correction but need redo search using higher ppm for unmatched mz
            result = list_search_formula_mass_db(list_query_mz, indexed_DB, 2*ppm_std)

    return {
        'mass_accuracy': mass_accuracy,
        'ppm_std': ppm_std,
        'annotated_list': result,
        'corrected_list_query_mz': corrected_list_query_mz,
        }





'''
Constructing empirical compounds de novo.

The ion patterns here are constructed differently from formula based calculations,
because formulae are not known here.

        Anchor peak is the most abundant isotopic peak. Modification peaks are from adducts, neutral loss and fragments.
        Requires at least one of isotopic_signatures and adduct_signatures. Use peak numbers only.
        If initiated by isotopic_signatures, do adduct search;
        elif initiated by adduct_signatures, assuming other primary isotopes are not seen.
        
        We build indexed centurion trees here to assist searches.
        Minor possibility that peak:empCpd is not N:1.

'''
from .search import *

class epdsConstructor:
    '''
    To organize a table of peaks/features into a list of empirical compounds
    (https://github.com/shuzhao-li/metDataModel).
    empCpds (or epds) can be initiated by signatures on isotopic relationships or adduct relationships.
    For low-intensity peaks, their isotopic counterparts may not be detectable.
    Only common adducts are considered at this step.
    This list of epds are used in m/z alignment/correspondence.
    For assigned peaks/features, will calculate selectivity/score later too.

    Input
    =====
    isotopic_patterns: [(1.003355, '13C/12C', (0, 0.2)), ...],
    peak_list: [{'parent_masstrace_id': 1670, 'mz': 133.09702315984987, 'apex': 654, 'height': 14388.0, 
    'left_base': 648, 'right_base': 655, 'id_number': 555},
                 ...]
    
    Example
    =======
    To get list of empCpds, run
    ECCON = epdsConstructor(list_peaks)
    list_empCpds = ECCON.peaks_to_epds()
    # e.g. {'id': 358, 'list_peaks': [(4215, 'anchor'), (4231, '13C/12C'), (4339, 'anchor,+NH4')]},

    To add -
    support of user input formats where rtime isn't precise or unavailable.

    '''
    def __init__(self, peak_list, mode='pos'):
        '''
        self.peak_dict: {peak_id: {'parent_masstrace_id': 1670, 'mz': 133.09702315984987, 'apex': 654, 'height': 14388.0, 
                    'left_base': 648, 'right_base': 655, 'id_number': 555}, ...}
        '''
        # will add parameters dict
        self.peak_list = peak_list
        self.peak_dict = build_peak_id_dict(self.peak_list)
        self.mode = mode
        self.seed_search_patterns = seed_empCpd_patterns[self.mode]
        self.ext_search_patterns = [x for x in common_adducts[self.mode] if x not in self.seed_search_patterns]

    def peaks_to_epdDict(self, 
                            seed_search_patterns, 
                            ext_search_patterns,
                            mz_tolerance_ppm=5, 
                            check_coelution=True,
                            isotope_rt_tolerance=5,     # default is scan number
                            coelution_rt_tolerance=10, 
                            coelution_function='overlap',
                            check_isotope_ratio = True):
        '''
        Wrapper function of constructing empCpds (empricial compounds).
        This complies with empCpd JSON convention, while self.peaks_to_epds produces short-handed lists.
        Input
        =====
        mz_tolerance_ppm: ppm tolerance in examining m/z patterns.
        seed_search_patterns: initial ions for constructing an empCpd. 
                This is very conservative and requires stricter isotope_rt_tolerance, because other ions can be searched later. 
                Can use self.seed_search_patterns, e.g. for pos ions:
                [(1.003355, '13C/12C', (0, 0.8)), (1.0078, 'H'), (21.9820, 'Na/H')]
        ext_search_patterns: 
                Additional ions to be searched to extend seed empCpds. Can use self.ext_search_patterns.
        check_coelution: True unless retention time is not available.
        isotope_rt_tolerance: tolerance threshold for deviation in retetion time, stricter version for seed ions (isotopes). 
                Default unit is scan number, but can be arbitrary.
        coelution_rt_tolerance: tolerance threshold for deviation in retetion time of adducts (other than seeds).
        coelution_function: default is 'overlap', checking overlap of peaks. Fallback is by distance, if any other value is given.

        Return
        ======
        list_empCpds, e.g. [{'interim_id': 12,
                        'neutral_formula_mass': None,
                        'neutral_formula': None,
                        'Database_referred': [],
                        'identity': [],
                        'MS1_pseudo_Spectra': [{'id_number': 'F94',
                        'mz': 98.97531935309317,
                        'apex': 700.0,
                        'ion_relation': 'anchor',
                        'parent_epd_id': 12},
                        {'id_number': 'F8161',
                        'mz': 99.9786892845517,
                        'apex': 701.0,
                        'ion_relation': '13C/12C',
                        'parent_epd_id': 12},
                        {'id_number': 'F317',
                        'mz': 116.98587698692174,
                        'apex': 700.0,
                        'ion_relation': 'anchor,+H2O',
                        'parent_epd_id': 12}],
                        'MS2_Spectra': []}, ...]
        '''
        return self.index_reformat_epds( 
                        self.peaks_to_epds(
                        seed_search_patterns, ext_search_patterns, mz_tolerance_ppm, 
                        check_coelution, isotope_rt_tolerance, coelution_rt_tolerance, coelution_function, check_isotope_ratio) 
                        )

    def peaks_to_epds(self, seed_search_patterns, ext_search_patterns,
                            mz_tolerance_ppm=5, 
                            check_coelution=True,       # use later
                            isotope_rt_tolerance=5, coelution_rt_tolerance=10, 
                            coelution_function='overlap',
                            check_isotope_ratio=True):
        '''
        Seeds by core set of 12C/13C isotopes and common adducts; then extend to common adducts.
        exhaustive_search is discouraged here, because it can be done better when formulae are supplied by a reference database.
        See parameter descriptions in peaks_to_epdDict.

        Return
        ======
        list_empCpds, [{'id': ii, 'list_peaks': [(peak_id, ion), (), ...],}, ...]
            The first peak is anchor ion.
        '''
        print("\n\nAnnotating empirical compounds on %d features/peaks, ..." %len(self.peak_list))
        if coelution_function == 'overlap':
            is_coeluted = is_coeluted_by_overlap
        else:
            is_coeluted = is_coeluted_by_distance
            # usage e.g. is_coeluted(P1, P2, coelution_rt_tolerance=10)

        list_empCpds = []
        epds = self.build_epds_2_steps(self.peak_list, seed_search_patterns, ext_search_patterns, 
                        mz_tolerance_ppm, 
                        isotope_rt_tolerance, coelution_rt_tolerance, is_coeluted, check_isotope_ratio)
        found2 = []
        for L in epds:
            found2 += [x[0] for x in L]
        found2 = set(found2)
        _NN2, ii = len(found2), -1
        for ii in range(len(epds)):
            list_empCpds.append(
                {'id': ii, 'list_peaks': epds[ii]}
            )
        print("epdsConstructor - numbers of seeded epds and included peaks: ", (ii+1, _NN2))
        return list_empCpds

    def build_epds_2_steps(self, peak_list, search_patterns, ext_search_patterns, mz_tolerance_ppm, 
                                isotope_rt_tolerance, coelution_rt_tolerance, is_coeluted, check_isotope_ratio):
        '''
        return list of epds by 2-step search: seeds first, then extended by another set of search pattern.
        search_patterns, e.g. [(1.003355, '13C/12C', (0, 0.8)), (10.991, 'Na/H, double charged'), (18.0106, '+H2O')]
        '''
        epds, found = [], []
        mztree = build_centurion_tree(peak_list)
        seeds = get_seed_empCpd_signatures(peak_list, mztree, search_patterns, mz_tolerance_ppm, 
                        isotope_rt_tolerance, coelution_rt_tolerance, is_coeluted, check_isotope_ratio)
        # [[(182, 'anchor'), (191, '13C/12C'), (205, 'H')], ...]
        for L in seeds:
            found += [x[0] for x in L]
        found = set(found)
        remaining_peaks = [P for P in peak_list if P['id_number'] not in found]
        mztree = build_centurion_tree(remaining_peaks)
        for G in seeds:
            epds.append(
                G + extend_seed_empCpd_signature(G, self.peak_dict, mztree, ext_search_patterns, 
                        mz_tolerance_ppm, coelution_rt_tolerance, is_coeluted)
            )
        return epds


    def index_reformat_epds(self, list_empCpds):
        '''
        Careful with format inconsistency
        '''
        new = {}
        for E in list_empCpds:
            features = []
            for peak in E['list_peaks']:
                self.peak_dict[peak[0]]['ion_relation'] = peak[1]
                features.append( self.peak_dict[peak[0]] )
            new[E['id']] = {
                'interim_id': E['id'], 
                'neutral_formula_mass': None,
                'neutral_formula': None,
                'Database_referred': [],
                'identity': [],
                'MS1_pseudo_Spectra': features,
                'MS2_Spectra': [],
                }

        return new


# -----------------------------------------------------------------------------
#
    def exhaustive_build(self, ii, list_empCpds):
        '''
        exhaustive_search=False, exclude_singletons=True
                # the now remaining are to be assigned as singletons
        if not exclude_singletons:
            remaining_peaks = [P for P in remaining_peaks if P['id_number'] not in found2]
            for P in remaining_peaks:
                ii += 1
                list_empCpds.append(
                    {'id': ii, 'list_peaks': [(P['id_number'], 'anchor')]}
                )

        '''
        pass

    def __extend_empCpds_by_adducts(self, seed_list_empCpds, list_peaks, adduct_patterns, mz_tolerance_ppm=5):
        '''
        Not used now.
        Search list_peaks for adducts that fit patterns relative to anchor ions in existing empCpds.
        Co-elution required.
        This is not done in initial empCpd construction to reduce errors in assigning peaks to empCpds.

        in progress -
        '''
        
        mztree = build_centurion_tree(list_peaks)
        for EPD in seed_list_empCpds:
            P1, relation = EPD['list_peaks'][0]
            for adduct in adduct_patterns:
                # (1.0078, 'H'), (21.9820, 'Na/H'), ...
                tmp = find_all_matches_centurion_indexed_list( P1['mz'] + adduct[0], mztree, mz_tolerance_ppm )
                for P2 in tmp:
                    if is_coeluted(P1, P2):
                        EPD['list_peaks'].append( (P2['id_number'], relation +','+ adduct[1]) )


        return seed_list_empCpds


    def __consolidate_epds(self, epd_peaks):
        '''
        Merge redundant empirical compounds (epds), which can arise when diff isotopes have same adducts,
        by 1) removing an epd if it's a subset of another epd;
           2) resolving each peak to only one epd.
        epd_peaks = [[(182, 'anchor'), (191, '13C/12C'), (205, '18O/16O')], ...]
        '''
        ordered_set_peak_ids = [set([x[0] for x in P]) for P in epd_peaks]
        def _choose(jj, ii, ordered_set_peak_ids=ordered_set_peak_ids, epd_peaks=epd_peaks):
            if ordered_set_peak_ids[ii].issubset(ordered_set_peak_ids[jj]):
                return jj
            elif ordered_set_peak_ids[jj].issubset(ordered_set_peak_ids[ii]):
                return ii
            else:           # warning for now; should choose the more common ion
                print("Unresolved peak between ", ii, jj)
                #
                # to update
                #
                return jj

        p2e = {}
        for ii in range(len(ordered_set_peak_ids)):
            for p in ordered_set_peak_ids[ii]:
                if p not in p2e:
                    p2e[p] = ii
                else:
                    p2e[p] = _choose(p2e[p], ii)

        valid_epds = set(p2e.values())
        # this does not completely resolve peak ownership
        return [epd_peaks[ii] for ii in valid_epds]



from collections import namedtuple
import numpy as np

# from metDataModel.core import Feature

# intensities is a list; mass_id links to MassTrace
tFeature = namedtuple('tFeature', ['feature_id', 'mass_id', 'mz', 'rtime', 'rt_min', 'rt_max', 
                                'peak_quality_max', 'peak_quality_median', 'number_peaks', 'perc_peaks',
                                'selectivity_combined', 'selectivity_mz', 'intensity_mean',
                                'intensities'])


def read_feature_tuples(feature_table, 
                        id_col=True, mz_col=1, rtime_col=2, 
                        intensity_cols=(3,4), delimiter="\t"):
    '''
    Read a text feature table into a list of features.

    Input
    -----
    feature_table: Tab delimited feature table. First line as header.
                    Recommended col 0 for ID, col 1 for m/z, col 2 for rtime.
    id_col: column for id. If feature ID is not given, row_number is used as ID.
    mz_col: column for m/z.
    rtime_col: column for retention time.
    intensity_cols: range of columns for intensity values.

    Return
    ------
    List of namedtuple 'Features': [(Feature)...], 
    '''
    featureLines = open(feature_table).read().splitlines()
    header = featureLines[0].split(delimiter)
    num_features = len(featureLines)-1
    # sanity check
    print("table headers ordered: ", header[mz_col], header[rtime_col])
    print("Read %d feature lines" %num_features)
    L = []
    for ii in range(1, num_features+1):
        if featureLines[ii].strip():
            a = featureLines[ii].split(delimiter)
            if isinstance(id_col, int):         # feature id specified
                iid = a[id_col]
            else:
                iid = 'row'+str(ii)
            xstart, xend = intensity_cols
            intensities = [float(x) for x in a[xstart: xend]]
            L.append({
                'id': iid, 'mz': float(a[mz_col]), 'rtime': float(a[rtime_col]),
                'intensities': intensities,
                'representative_intensity': np.mean(intensities),
            })
    return L



def index_features(list_of_features, range_low=20, range_high=2000):
    '''
    Indexing a list of features to speed up search.
    A filter/Exception should be added for the mass range in production.
    Internal use.

    Input
    -----
    List of features, [{'id': '', 'mz': 0, 'rtime': 0, ...}, ...]
    range_low, range_high: m/z range, default [20, 2000]. 

    Return
    ------
    Dictionary of features, indexed using integer of m/z.
    '''
    L = {}
    for ii in range(range_low, range_high+1): L[ii] = []
    for F in list_of_features: L[int(F['mz'])].append(F)
    return L


def read_features(feature_table, 
                        id_col=True, mz_col=1, rtime_col=2, 
                        intensity_cols=(3,4), delimiter="\t"):
    '''
    Read a text feature table into a list of features.
    Internal use.

    Input
    -----
    feature_table: Tab delimited feature table. First line as header.
                    Recommended col 0 for ID, col 1 for m/z, col 2 for rtime.
    id_col: column for id. If feature ID is not given, row_number is used as ID.
    mz_col: column for m/z.
    rtime_col: column for retention time.
    intensity_cols: range of columns for intensity values.

    Return
    ------
    List of features: [{'id': '', 'mz': 0, 'rtime': 0, 
                        intensities: [], 'representative_intensity': 0, ...}, 
                        ...], 
                        where representative_intensity is mean value.
    '''
    featureLines = open(feature_table).read().splitlines()
    header = featureLines[0].split(delimiter)
    num_features = len(featureLines)-1
    # sanity check
    print("table headers ordered: ", header[mz_col], header[rtime_col])
    print("Read %d feature lines" %num_features)
    L = []
    for ii in range(1, num_features+1):
        if featureLines[ii].strip():
            a = featureLines[ii].split(delimiter)
            if isinstance(id_col, int):         # feature id specified
                iid = a[id_col]
            else:
                iid = 'row'+str(ii)
            xstart, xend = intensity_cols
            intensities = [float(x) for x in a[xstart: xend]]
            L.append({
                'id': iid, 'mz': float(a[mz_col]), 'rtime': float(a[rtime_col]),
                'intensities': intensities,
                'representative_intensity': np.mean(intensities),
            })
    return L



def text_to_mcg_features(textValue, MASS_RANGE = (50, 2500), delimiter='\t'):
    '''
    Parse text input to List of features, in JSON peak formats. 
    The older versions used instances of metDataModel.Feature, which is better for complex operations.
    Here, simplicity is preferred.

    Input
    -----
    Text string of feature table, mummichog format, Column order is hard coded for now, 
    1st row as header = [mz, retention_time, p_value, statistic, CompoundID_from_user].
    Each row is a unique feature. Redundant entries should be check prior to this function.

    Return
    -------
    peak_list: [{''id_number': 555, 'mz': 133.09702315984987, 'apex': 654, 
    'height': 14388.0, 
    'left_base': 648, 'right_base': 655, },
                 ...]
    list of excluded lines in input text
    '''
    list_of_features, excluded_list = [], []
    lines = textValue.splitlines()
    _header_fields = lines[0].rstrip().split(delimiter)

    for ii in range(len( lines )-1):
        y = lines[ii+1].split('\t')
        CompoundID_from_user = ''
        if len(y) > 4: CompoundID_from_user = y[4]
        [mz, retention_time, p_value, statistic] = [float(x) for x in y[:4]]
        
        if MASS_RANGE[0] < mz < MASS_RANGE[1]:
            # row # human-friendly, numbering from 1
            F = {
                'id_number': 'row'+str(ii+1),
                'mz': mz,
                'retention_time': retention_time,
                'apex': retention_time,
                'statistic_score': statistic,
                'p_value': p_value,
                'CompoundID_from_user': CompoundID_from_user,
            }
            list_of_features.append( F )
        else:
            excluded_list.append( lines[ii+1] )

    return (list_of_features, excluded_list)


def textfile_to_mcg_features(infile):
    return text_to_mcg_features(open(infile).read())




'''
Two lines of search methods:
- A centurion_tree is an indexed dictionary of peaks/features.
- We also use DataFrame based vector operations for some cases (less often).

Emperical compounds are constructed by co-eluting isotopic and adduct patterns.
# isotopic_signatures: example as [(182, 'anchor'), (191, 'M(13C)'), (205, 'M(18O)')]
# consider Cl as adduct, both 35Cl and 37Cl are considered
# 34S and 37Cl are close, but should have diff ratios and are formula dependent

'15N/14N' and '33S/32S' are close for high m/z features, e.g. 0.002353/470 ~ 5 pmm.
'18O/16O' and 'M(13C),M(15N)' are close for high m/z features, e.g. 0.003855/770 ~ 5 pmm.

Comprehensive combinations of isotopes and adducts, vs seeded search:
all adducts can be combined with isotopes in theory, greatly depending on abundance in practice.
Controlled by coelution, it's high confidence in assigning these peaks to empCpds, 
but not 100% correct. 
Until we have a better strategy, step-wise searches are less prone to errors.

search_patterns: Use seed_empCpd_patterns. May update after seeing experimental statistics.
                    Avoid H2O because hard to distinguish from differences in chemical formulae.
                    The third item in tuples is to control abundance ratio, 
                    which is linient to leave room for high carbon numbers and measurement.

'''

# Mass difference and Abundance of natually occuring isotopic elements.
# (mz difference, notion, ratio low limit, ratio high limit)
# Ratio is loose to leave room for 1) multiple atoms, e.g. 40 C atoms lead to ~ 40*1% in abundance
# 2) measurement errors; not all peaks are in linear dynamic range
# 3) quantification by peak height is not same as peak area

isotopic_patterns = [
    # mass diff, isotopes, (intensity ratio constraint)
    (1.003355, '13C/12C', (0, 0.8)),      # 13C-12C, 12C~99%, 13C ~ 1%
    (0.997035, '15N/14N', (0, 0.2)),     # 15N-14N, 14N ~ 99.64%, 15N ~ 0.36%
    (2.004245, '18O/16O', (0, 0.2)),      # 18O-16O, 16O ~ 99.76, 16O ~ 0.2%
    (1.995796, '34S/32S', (0, 0.4)),      # 32S (95.02%), 33S (0.75%), 34S (4.21%)
    (0.999388, '33S/32S', (0, 0.1)),
    # double isotopes
    (2.00039, 'M(13C),M(15N)', (0, 0.2)),
    (2.999151, 'M(13C),M(34S)', (0, 0.4)),
    # double charged
    (0.5017, '13C/12C, double charged', (0, 0.8)),
    (0.4985, '15N/14N, double charged', (0, 0.2)),
]

common_adducts = {
    # mass diff, modification
    # not using (intensity ratio constraint), but it can be documented or learned
    'pos': [
        (1.0078, 'H'),
        (21.9820, 'Na/H'), # Na replacing H
        (10.991, 'Na/H, double charged'),
        (18.0106, '+H2O'), 
        (18.033823, '+NH4'),
        (37.9559, '39K/H'),
        (39.9540, '41K/H'),
        (41.026549, 'Acetonitrile'),
    ],
    'neg': [
        (1.0078, 'H'),
        (22.9893, 'Na'),
        (20.97474706646, '+Na-2H'),
        (18.0106, 'H2O'), 
        (34.9689, '35Cl'),
        (36.9659, '37Cl'),
        (40.01926853323, '+ACN-H'),
        (44.998201, 'COOH'),
        (59.013295, 'CH3COO'),
    ],
}

extended_adducts = {
    'pos': [],
    'neg': [],
}

# this does not include combinatorial values of isotopes and adducts
# Mostly used as seeds; better to do inclusive search after matched to formulae
seed_empCpd_patterns = {
    'pos': [(1.003355, '13C/12C', (0, 0.8)), (1.0078, 'H'), (21.9820, 'Na/H'), ],
    'neg': [(1.003355, '13C/12C', (0, 0.8)), (1.0078, 'H'), (20.97474706646, '+Na-2H'), (34.9689, '35Cl')],
}



#
# -----------------------------------------------------------------------------
#

def build_centurion_tree(list_peaks):
    '''
    list_peaks: [{'parent_masstrace_id': 1670, 'mz': 133.09702315984987, 'apex': 654, 'height': 14388.0, 
                    'left_base': 648, 'right_base': 655, 'id_number': 555}, ...]
    Return a dictionary, indexing mzList by 100*mz bins.
    Because most high-resolution mass spectrometers measure well under 0.01 amu, 
    one only needs to search the corresponding 0.01 bin and two adjacent bins (to capture bordering values).
    list_mass_tracks has similar format as list_peaks.
    '''
    d = {}
    for p in list_peaks:
        cent = int(100 * p['mz'])
        if cent in d:
            d[cent].append(p)
        else:
            d[cent] = [p]
    return d


def build_peak_id_dict(list_peaks):
    d = {}
    for p in list_peaks:
        d[p['id_number']] = p
    return d

def __build_centurion_tree_mzlist(mzList):
    '''
    Return a dictionary, indexing mzList by 100*mz bins.
    Because most high-resolution mass spectrometers measure well under 0.01 amu, 
    one only needs to search the corresponding 0.01 bin and two adjacent bins (to capture bordering values).
    '''
    d = {}
    for ii in range(len(mzList)):
        cent = int(100*mzList[ii])
        if cent in d:
            d[cent].append((mzList[ii], ii))
        else:
            d[cent] = [(mzList[ii], ii)]
    return d

def find_all_matches_centurion_indexed_list(query_mz, mz_centurion_tree, limit_ppm=5):
    '''
    Return matched peaks in mz_centurion_tree.
    '''
    q = int(query_mz * 100)
    mz_tol = query_mz * limit_ppm * 0.000001
    results = []
    for ii in (q-1, q, q+1):
        L = mz_centurion_tree.get(ii, [])
        for peak in L:
            if abs(peak['mz']-query_mz) < mz_tol:
                results.append(peak)
                
    return results

def find_best_match_centurion_indexed_list(query_mz, mz_centurion_tree, limit_ppm=2):
    '''
    Return matched indices in mz_centurion_tree (based on peak list).
    '''
    q = int(query_mz * 100)
    mz_tol = query_mz * limit_ppm * 0.000001
    result = (None, 999)
    for ii in (q-1, q, q+1):
        L = mz_centurion_tree.get(ii, [])
        for peak in L:
            _d = abs(peak['mz']-query_mz)
            if _d < min(result[1], mz_tol):     # enforce mz_tol here
                result = (peak, _d)
                
    return result[0]

def is_coeluted_by_overlap(P1, P2, rt_tolerance=10):
    '''
    coelution is defined by verlap more than half of the smaller peak, or apexes within rt_tolerance.
    If not enough parameters are given for peaks, fallback to is_coeluted_by_distance(P1, P2, rt_tolerance=10).
    
    Example peak format: {'parent_masstrace_id': 1670, 'mz': 133.09702315984987, 'apex': 654, 'height': 14388.0,
    'left_base': 648, 'right_base': 655, 'id_number': 555}

    rt_tolerance: tolerance of retention time is used when peak boundaries are not given.
                Can be scan numbers or seconds.
                
    return: True or False
    '''
    _coeluted = False
    try:
        len1, len2 = P1['right_base'] - P1['left_base'], P2['right_base'] - P2['left_base']
        # overlap is the max L and min R
        overlap = min(P1['right_base'], P2['right_base']) - max(P1['left_base'], P2['left_base'])
        if overlap > 0.5 * min(len1, len2):
            _coeluted = True
        return _coeluted
    except:
        print("KeyError, fallback to function is_coeluted_by_distance.")
        return is_coeluted_by_distance(P1, P2, rt_tolerance=10)

def is_coeluted_by_distance(P1, P2, rt_tolerance=10):
    _coeluted = False
    if abs(P1['apex']-P2['apex']) <= rt_tolerance:
        _coeluted = True
    return _coeluted

def get_seed_empCpd_signatures(list_peaks, 
                    mztree, 
                    search_patterns = seed_empCpd_patterns['pos'],
                    mz_tolerance_ppm=5, 
                    isotope_rt_tolerance=5, coelution_rt_tolerance=10, 
                    is_coeluted = is_coeluted_by_overlap,
                    check_isotope_ratio = True,
                    ):
    '''
    To find a core group of ions that belong to an empirical compound (emp_cpd, or empCpd, ecd).
    This should be a combination of 12C/13C isotopes and common adducts;
    an empCpd can be constructed even if the 13C ion is missing.
    Other isotopes and adducts can be searched after establishing the seed empCpds.

    This is currently a heuristic approach to generate empCpds without too many errors.
    One may research more to polish an optimal method of assigning ions to empCpds, 
    where chemical formulae and learned patterns can be incorporated.

    Input
    =====
    list_peaks: [{'parent_masstrace_id': 1670, 'mz': 133.09702315984987, 'apex': 654, 'height': 14388.0, 
                    'left_base': 648, 'right_base': 655, 'id_number': 555}, ...]
    mztree: indexed list_peaks
    mz_tolerance_ppm: ppm tolerance in examining m/z patterns.
    seed_search_patterns: initial ions for constructing an empCpd. 
            This is very conservative and requires stricter isotope_rt_tolerance, because other ions can be searched later. 
            Can use self.seed_search_patterns, e.g. for pos ions:
            [(1.003355, '13C/12C', (0, 0.8)), (1.0078, 'H'), (21.9820, 'Na/H')]
    isotope_rt_tolerance: tolerance threshold for deviation in retetion time, stricter version for seed ions (isotopes). 
            Default unit is scan number, but can be arbitrary.
    coelution_rt_tolerance: tolerance threshold for deviation in retetion time of adducts (other than seeds).
    is_coeluted: coelution function

    Return
    ======
    list of lists of peak numbers that match search_patterns patterns, e.g.
    [ [(195, 'anchor'), (206, '13C/12C')],  ...]
    '''
    signatures = []
    for P1 in list_peaks:
        matched = [ (P1['id_number'], 'anchor'), ]          # Nature is nice to have lowest mass for the most abundant 
        for _pair in search_patterns:
            (mass_difference, relation) = _pair[:2]
            tmp = find_all_matches_centurion_indexed_list(P1['mz'] + mass_difference, mztree, mz_tolerance_ppm)
            for P2 in tmp:
                if abs(P1['apex']-P2['apex']) <= isotope_rt_tolerance and is_coeluted(P1, P2, coelution_rt_tolerance):
                    if check_isotope_ratio and len(_pair) > 2:  # checking abundance ratio
                        (abundance_ratio_min, abundance_ratio_max) = _pair[2]
                        if abundance_ratio_min*P1['height'] < P2['height'] < abundance_ratio_max*P1['height']:
                            matched.append( (P2['id_number'], relation) )
                    else:
                        matched.append( (P2['id_number'], relation) )
 
        if len(matched) > 1:
            # if len(matched) > 2: matched = merge_confused_peak_matches(matched)
            signatures.append(matched)

    return signatures


def extend_seed_empCpd_signature(seed_signatures, peak_dict, mztree, ext_search_patterns, 
                    mz_tolerance_ppm=5, 
                    coelution_rt_tolerance=10, 
                    is_coeluted = is_coeluted_by_overlap,
                    ):
    '''
    Extend a single signature by ext_search_patterns, including combinatorial values of isotopes and adducts.

    Input
    =====
    mztree: indexed list_peaks
    seed_signatures: e.g. [(182, 'anchor'), (191, '13C/12C'), (205, '18O/16O')]
    search_patterns: e.g. [(10.991, 'Na/H, double charged'), (18.0106, '+H2O'), (18.033823, '+NH4'),]
    example peak: {'parent_masstrace_id': 1670, 'mz': 133.09702315984987, 'apex': 654, 'height': 14388.0, 
                    'left_base': 648, 'right_base': 655, 'id_number': 555}

    Return
    ======
    One epd (empirical compound) based on seed peaks and ext_search_patterns, as a list of [(peak_id, relation), ...]
    '''
    matched = []
    for seed in seed_signatures:
        peak, relation = peak_dict[seed[0]], seed[1]
        for adduct in ext_search_patterns:
            tmp = find_all_matches_centurion_indexed_list( peak['mz'] + adduct[0], mztree, mz_tolerance_ppm )
            for P2 in tmp:
                if is_coeluted(peak, P2, coelution_rt_tolerance):
                    matched.append( (P2['id_number'], relation +','+ adduct[1]) )

    return matched


def find_isotopic_signatures(list_peaks, mztree, isotopic_patterns, mz_tolerance_ppm=5, rt_tolerance_scans=5):
    '''
    See find_isotopic_pairs. This extends to all related isotopic signatures.
    Allows ambiguous matches, which will be dealt with in empCpd statistics.
    But the peaks need to be unique. E.g. 
    '15N/14N' and '33S/32S', '18O/16O' and 'M(13C),M(15N)' are close for high m/z features.
    Matches like ('F5486', '15N/14N'), ('F5486', '33S/32S') need to merge on the peak/feature.

    Input
    =====
    isotopic_patterns = [(1.003355, '13C/12C', (0, 0.8)), ...], the third item is optional limits of abundance ratio.

    Return
    ======
    list of lists of peak numbers that match isotopic patterns, [ [], ... ]

    Example
    =======
    [ [(195, 'anchor'), (206, '13C/12C')], 
      [(182, 'anchor'), (191, '13C/12C'), (205, '18O/16O')],
      [(295, 'anchor'), (335, '13C/12C'), (368, 'M(13C),M(34S)')], ...]
    '''
    signatures = []
    for P1 in list_peaks:
        matched = [ (P1['id_number'], 'anchor'), ]          # Nature is nice to have lowest mass for the most abundant 
        for isotopic_pair in isotopic_patterns:
            (mass_difference, relation) = isotopic_pair[:2]
            tmp = find_all_matches_centurion_indexed_list(P1['mz'] + mass_difference, mztree, mz_tolerance_ppm)
            for P2 in tmp:
                if abs(P1['apex']-P2['apex']) <= rt_tolerance_scans and is_coeluted_by_overlap(P1, P2):
                    if len(isotopic_pair) == 2:             # not checking abundance ratio
                        matched.append( (P2['id_number'], relation) )
                    else:
                        (abundance_ratio_min, abundance_ratio_max) = isotopic_pair[2]
                        # checking abundance ratio
                        if abundance_ratio_min*P1['height'] < P2['height'] < abundance_ratio_max*P1['height']:
                            matched.append( (P2['id_number'], relation) )
        if len(matched) > 1:
            # if len(matched) > 2: matched = merge_confused_peak_matches(matched)
            signatures.append(matched)

    return signatures


def find_adduct_signatures(list_peaks, mztree, adduct_patterns, mz_tolerance_ppm=5):
    '''
    Search adduct mass_diff in ceelution peaks, de novo. 
    Not requiring matched apex as in isotopic pairs, nor abundance ratio restriction.
    '''
    signatures = []
    for P1 in list_peaks:
        matched = [ (P1['id_number'], 'anchor'), ]
        for adduct in adduct_patterns:
            tmp = find_all_matches_centurion_indexed_list( P1['mz'] + adduct[0], mztree, mz_tolerance_ppm )
            for P2 in tmp:
                if is_coeluted_by_overlap(P1, P2):
                    matched.append( (P2['id_number'], adduct[1]) )

        if len(matched) > 1:
            signatures.append(matched)

    return signatures


def merge_confused_peak_matches(L):
    '''
    merge if same peak is matched to more than one isotopes or adducts.
    L example: [(295, 'anchor'), (335, '13C/12C'), (368, 'M(13C),M(34S)')]
    '''
    tmp, _d = [L[0]], {}
    for peak in L[1:]:
        if peak[0] in _d:
            _d[peak[0]] += ';' + peak[1]
        else:
            _d[peak[0]] = peak[1]
    for k,v in _d.items():
        tmp.append((k,v))
    return tmp


def find_mzdiff_pairs_from_masstracks(list_mass_tracks, list_mz_diff=[1.003355, 21.9820], mz_tolerance_ppm=5):
    '''
    Find all pairs in list_mass_tracks that match a pattern in list_mz_diff, and return their id_numbers as pairs.
    This function does not use coeluction (rtime) rules. 

    Input
    =====
    list_mass_tracks: [{ 'id_number': ii,  'mz': xic[0], 'rt_scan_numbers': xic[1],  'intensity': xic[2],  }, ...]
    list_mz_diff: defaul 1.003355, 21.9820 are 13C/12C, Na/H. Dependent on ionization mode.
    
    Return
    ======
    list of pairs of mass tracks numbers.
    '''
    pairs = []
    # list_mass_tracks has similar format as list_peaks.
    mztree = build_centurion_tree(list_mass_tracks)
    for mzdiff in list_mz_diff:
        for P1 in list_mass_tracks:
            P2 = find_best_match_centurion_indexed_list(P1['mz'] + mzdiff, mztree, mz_tolerance_ppm)
            if P2:
                pairs.append((P1['id_number'], P2['id_number']))

    return pairs


#
# -----------------------------------------------------------------------------
#

def search_formula_mass_dataframe(query_mz, DFDB, limit_ppm=10):
    '''
    return best match formula_mass in DFDB if under ppm limit.
    DFDB is using a Pandas DataFrame to house reference database.
    ppm is signed to capture the direction of mass shift.

    Not using idxmin,   #ii = DFDB.tmp.idxmin()
    # in pd.DF, index not necessarily integer; can be sequence if more than one match, but they are trying to fix in pandas dev version
    '''
    DFDB['tmp'] = abs(DFDB.mz - query_mz)
    ii = DFDB.tmp.values.argmin()
    ppm = 1000000 * (query_mz - DFDB.iloc[ii].mz)/query_mz
    if  abs(ppm) < limit_ppm:
        return (DFDB.iloc[ii].name, ppm)            # name is formula_mass
    else:
        return None




'''
Computational calibration of meassured masses,
assuming that a common set of metabolites exist in the samples.
The chance of matching the accurate mass of large number of ions is unambiguous.

Multiple methods to calibrate:

    1. A regression model is built for (theoretical m/z) = a * (measured m/z) + b + e,
    then the model is applied to all features to calculate a calibrated mz.
    Because the model is too sensitive to `a` and mass shift in this slope is often negligible,
    method 2 is preferred.

    2. (measured m/z)/(theoretical m/z) - 1 =  b + e. Systematic shift can be reflected in `b`.
    b*10e6 is conventional ppm.
    Assuming a normal distribution of mass errors, b and std are fitted.

    3. (theoretical m/z) = b + e*f(measured m/z). The accuracy is a function of m/z, as 
    larger m/z has larger errors. The convention is to use a dalton window (0 order) or ppm (1st order),
    but Orbitrap data suggest that we need a 2nd order model. 

We implement method 2 for now. Method 3 will be on to-do list.

Suggested use:
from .calibrate import normal_error_distribution as mass_calibrate


These have to be based on high-selectivity ions, otherwise ground truth is hard to establish,
and the statistics is not valid.

This should be done after correspondence of high-selectivity features 
(after peak detection and initial annotation).
If the mass calibration is off, correct shift (mu) and use empirical stdev to recalculate selectivity.



'''

from scipy.stats import norm

from .io import index_features

def mass_calibrate(mz_List_1, mz_List_2):
    '''
    mz_List_1, mz_List_2 as numpy arrays.
    return
    ------
    The mean (List2 - List1) and standard deviation of mass differences. 
    The latter can be used as a guide of precision in an experiment.

    Example
    -------
    p = norm.pdf(xx, mu, std)
    plt.plot(xx, p, 'kx', linewidth=2)

    '''
    return norm.fit(mz_List_2 - mz_List_1)



# -----------------------------------------------------------------------------
#
# Prototype code with short-handed data structure
#
# -----------------------------------------------------------------------------

def old_mass_calibrate(query_features, anchor_features, limit_ppm=10):
    '''
    Use a list of common metabolites to calibrate m/z values in query_features 
    (e.g. all m/z values in a smaple).
    Assuming a large number of compounds in ref_mass_dict exist in the list_of_features,
    which should be the case in common biological matrices.
    This function does not require complete use of anchor_features.
    The mass errors are assumed to fit a normal distribution, 
    and it's parameters (e.g. 2*stdev) are used for subsequent confidence in matching.
    
    Input
    -----
    query_features: List of m/z values from the same Experiment.
                
    anchor_features: list of m/z values for metabolites known to be in the sample.
                    They need not to be 100% present, but in a sufficient number 
                    to estimate error distribution.

    limit_ppm: default 10. Expecting a high-res instrument performs within 25 ppm, 

    Return
    ------
    coefficients: (mean, std) in the fitted model of normal distribution of mass errors.
    updated_query_features: List of updated m/z values for the query features in same order.
    '''
    matched = []
    for m in calibration_mass_dict.values():
        best_pair = _find_closest(m, indexed_features, limit_ppm)
        if best_pair:
            matched.append(best_pair)

    def _find_closest(query_mz, indexed_features, limit_ppm):
        # to find closest match of a theoretical m/z in indexed_features, under limit_ppm
        # return (delta, feature_mz, query_mz) or None
        _delta = query_mz * limit_ppm * 0.000001
        _low_lim, _high_lim = query_mz - _delta, query_mz + _delta
        result = []
        for ii in range(int(_low_lim), int(_high_lim)+1):
            for F in indexed_features[ii]:
                result.append( (abs(query_mz-F['mz']), F['mz'], query_mz) )

        result.sort()
        if result and result[0][0] < _delta:
            return result[0]
        else:
            return None

    def _objective(x, a, b): return a * x + b

    indexed_features = index_features(list_features)
    matched = []
    for m in calibration_mass_dict.values():
        best_pair = _find_closest(m, indexed_features, limit_ppm)
        if best_pair:
            matched.append(best_pair)

    num_matched = len(matched)
    if num_matched < 3:
        print("Aborted. Found %d matched mass features." %num_matched)

    else:
        if num_matched < 10:
            print("Warning. Using %d matched mass features, too few for regression." %num_matched)
        X, Y = [x[1] for x in matched], [x[2] for x in matched]
        popt, _ = curve_fit(_objective, X, Y)
        a, b = popt
        if abs(b) > 0.01 or abs(a-1) > 0.000025:
            print(" -- WARNING -- extreme parameters are usually bad, ", a, b)

        for F in list_features:
            F['calibrated_mz'] = a * F['mz'] + b

        return (a,b), list_features


def fake_mass_calibrate(list_features):
    for F in list_features:
        F['calibrated_mz'] = F['mz']
    return list_features



'''
Handling chemical formulae.

This should switch to a more dedicated chemical tool, 
handling structural formats too.

Primary goal in the Mummichog suite is to compute:
1. associations btw chemical formulae and masses
2. Isotopes
3. In source reactions, including adducts, neutral loss, fragments
4. Biochemical reactions

'''

import re
from collections import namedtuple

PROTON = 1.00727646677
electron = 0.000549

Ion = namedtuple('Ion', ['mz', 'ion', 'delta_formula'])


def parse_chemformula_dict(x):
    '''This does not deal with nested groups in chemical formula, or isotopes.
    Formula from HMDB 3.5 are compatible.

    Example
    -------
    parse_chemformula_dict('C3H6N2O2')
    {'C': 3, 'H': 6, 'N': 2, 'O': 2}
    '''
    p = re.findall(r'([A-Z][a-z]*)(\d*)', x)
    d = {}
    for pair in p: d[pair[0]] = int( pair[1] or 1 )
    return d


def add_formula_dict(dict1, dict2):
    '''
    Addition of two formulae as dictionaries.
    This allows calculating formula after a reaction, as dict2 can contain substraction of elements.
    Not as good as using real chemical structures, just a simple approximation.

    '''
    new, result = {}, {}
    for k in set(dict1.keys()).union( set(dict2.keys()) ):
        if k in dict1 and k in dict2:
            new[k] = dict1[k] + dict2[k]
        elif k in dict1:
            new[k] = dict1[k]
        else:
            new[k] = dict2[k]
    # check validity, as no negative number of element is allow in product
    _VALID = True
    for k in new:
        if new[k] < 0:
            _VALID = False
        elif new[k] > 0:     # remove if element is 0
            result[k] = new[k] 
    if _VALID:
        return result
    else:
        return None


def check_elemental_subset(Fm1, Fm2):
    '''
    Check if Fm2 a subunit of Fm1; Dictionary representation of chem formula.
    Not used now.
    '''
    Fm2_in_Fm1 = True
    Elements1 = Fm1.keys()
    Elements2 = Fm2.keys()
    if [x for x in Elements2 if x not in Elements1]:    #element not in Fm1
        return False
    else:
        for e in Elements2:
            if Fm2[e] > Fm1[e]: Fm2_in_Fm1 = False
        return Fm2_in_Fm1


def dict_to_hill_formula(d):
    '''
    Generate simple formula from dictionary.
    Hill notion: First C, H and other elements in alphabetical order.
    Carbon isotope (C13) is considered, but not other isotopic elements for now.

    Input 
    -----
    formula dictionary, e.g. {'C': 3, 'H': 6, 'N': 2, 'O': 2} 

    Return
    ------
    Formula in Hill notion, e.g. 'C3H6N2O2'

    '''
    def _str_element_number(int_x):
        # convert int_x to str and '' for int_x is 1. Not checking validity. 
        if int_x == 1:
            return ''
        else:
            return str(int_x)

    elements = list(d.keys())
    elements.sort()
    new_order = [x for x in elements if x not in ['C', 'H', '(C13)']]
    if 'H' in d:
        new_order = ['H'] + new_order
    if '(C13)' in d:
        new_order = ['(C13)'] + new_order
    if 'C' in d:
        new_order = ['C'] + new_order

    formula = ''
    for x in new_order:
        formula += x + _str_element_number(d[x])
    return formula


def __get_adduct_list__(mw, mode, primary_only):
    '''
    Return list of adduct or isotope, [(mz, ion, formula_change_dictionary), ...].
    Next step compute_adducts_formulae will validate the ion.
    Enforcing primary ions because others are not valid without them;
    but logic implemented in search functions, not here.
    '''

    if mode == 'pos': 
        primaryList = [
            (mw - electron, 'M[1+]', {}),                              # e not changing formula
            (mw + PROTON, 'M+H[1+]', {'H': 1}),                        #  
            (mw + 21.9820 + PROTON, 'M+Na[1+]', {'Na': 1}),            # Na = 21.9820 + PROTON = 22.9893
            (mw + 18.0106 + PROTON, 'M+H2O+H[1+]', {'H': 3, 'O':1}),   #  
            (mw + 18.033823, 'M+NH4[1+]', {'H': 4, 'N': 1}),
            ]
        if primary_only:
            return [Ion(*L) for L in primaryList]

        else:
            extendedList = [
                # Use {'C':-1} to force check that C exists in formula
                (mw +1.0034 - electron, 'M(C13)[1+]', {'C':-1, '(C13)':1}), 
                (mw +1.0034 + PROTON, 'M(C13)+H[1+]', {'C':-1, '(C13)':1, 'H':1}),
                
                (mw/2 + PROTON, 'M+2H[2+]', {'H': 2}),
                (mw/3 + PROTON, 'M+3H[3+]', {'H': 3}),

                (mw/2 + 0.5017 + PROTON, 'M(C13)+2H[2+]', {'C':-1, '(C13)':1, 'H':2}),
                (mw/3 + 0.3344 + PROTON, 'M(C13)+3H[3+]', {'C':-1, '(C13)':1, 'H':3}),
                # Sulfur:  32S (95.02%), 33S (0.75%), 34S (4.21%), .
                (mw +1.9958 + PROTON, 'M(S34)+H[1+]', {'S':-1, '(S34)':1, 'H':1}),
                # Chlorine has two stable isotopes, 35Cl (75.77%) and 37Cl (24.23%)
                (mw +1.9972 + PROTON, 'M(Cl37)+H[1+]', {'Cl':-1, '(Cl37)':1, 'H':1}),
                
                (mw/2 + 10.991 + PROTON, 'M+H+Na[2+]', {'H':1, 'Na':1}),
                (mw + 37.9555 + PROTON, 'M+K[1+]', {'K':1}),         # K = 37.9555 + PROTON = 38.9628
                (mw + 57.9586 + PROTON, 'M+NaCl[1+]', {'Na': 1, 'Cl': 1}), 
                
                (mw - 18.0106 + PROTON, 'M-H2O+H[1+]', {'H': -1, 'O': -1}), 
                (mw - 36.0212 + PROTON, 'M-H4O2+H[1+]', {'H': -3, 'O': -2}),
                (mw - 17.0265 + PROTON, 'M-NH3+H[1+]', {'H': -2, 'N': -1}),
                (mw - 27.9950 + PROTON, 'M-CO+H[1+]', {'H': 1, 'C': -1, 'O': -1}),
                (mw - 43.9898 + PROTON, 'M-CO2+H[1+]', {'H': 1, 'C': -1, 'O': -2}),
                (mw - 46.0054 + PROTON, 'M-HCOOH+H[1+]', {'H': -1, 'C': -1, 'O': -2}),
                (mw + 67.9874 + PROTON, 'M+HCOONa[1+]', {'H': 1, 'C': 1, 'O': 2, 'Na': 1}),
                (mw - 72.0211 + PROTON, 'M-C3H4O2+H[1+]', {'H': -3, 'C': -3, 'O': -2}),
                (mw + 83.9613 + PROTON, 'M+HCOOK[1+]', {'H': 1, 'C': 1, 'O': 2, 'K': 1}),

                # removed, too infrequent
                # (mw - 67.9874 + PROTON, 'M-HCOONa+H[1+]', 'HCO2Na'),
                # (mw - 83.9613 + PROTON, 'M-HCOOK+H[1+]', 'HCO2K'),
                ]
            return [Ion(*L) for L in primaryList + extendedList]

    elif mode == 'neg':
        primaryList = [
            (mw - PROTON, 'M-H[-]', {'H': -1}),       
            (mw + electron, 'M[-]', {}), 
            (mw - 18.0106 - PROTON, 'M-H2O-H[-]', {'H': -3, 'O':-1}),   
            (mw + 34.9689, 'M+Cl[-]', {'Cl':1}),
            ]
        if primary_only:
            return [Ion(*L) for L in primaryList]

        else:
            extendedList = [
                (mw + 1.0034 - PROTON, 'M(C13)-H[-]', {'C':-1, '(C13)':1, 'H':-1}),
                (mw + 36.9659, 'M+Cl37[-]', {'(Cl37)':1}),
                (mw + 1.9972 - PROTON, 'M(Cl37)-H[-]', {'Cl':-1, '(Cl37)':1, 'H':-1}),
                (mw + 1.9958 - PROTON, 'M(S34)-H[-]', {'S':-1, '(S34)':1, 'H':-1}),

                (mw/2 - PROTON, 'M-2H[2-]', {'H': -2}),
                (mw + 22.9893 - 2*PROTON, 'M+Na-2H[-]', {'H':-2, 'Na':1}),
                (mw + 38.9628 - 2*PROTON, 'M+K-2H[-]', {'H':-2, 'K':1}),
                
                (mw + 78.9183, 'M+Br[-]', {'Br':1}),
                (mw + 80.9163, 'M+Br81[-]', {'(Br81)':1}),

                (mw + 40.01926853323, 'M+ACN-H[-]', {'C':2, 'H':2, 'N':1}),
                (mw + 1.007825 + 12 + 2*15.99491, 'M+HCOO[-]', {'C':1, 'H':1, 'O':2}),
                (mw + 59.013295, 'M+CH3COO[-]', {'C':2, 'H':3, 'O':2}),
                (mw - PROTON + 15.99491, 'M-H+O[-]', {'H': -1, 'O':1}),
                ]
            return [Ion(*L) for L in primaryList + extendedList]


def compute_adducts_formulae(mw, neutral_formula,  mode='pos', primary_only=False):
    '''
    Calculating isotopes and adducts, return m/z and formulae.

    A dictionary of element changes is used in the calculation.
    Another function add_formula_dict checks validity of required numbers of atoms.
    Electron is not considered in formula calculation, but may affected mass calculation.

    The isotopes, neutral losses and adducts can have multiplications,
    but multiple rounds of search will take care of those.

    Reaction notion in dictionary: 
    {'C': 3, 'H': 6, 'N': 2, 'O': 2} changed by {'C': -1, 'H': 1, 'O': -1} 
        = {'C': 2, 'H': 5, 'N': 2, 'O': 1}

    Return
    -------
    List of adducts: e.g. [(58.53894096677, 'M+2H[2+]', result_formula), ...,]

    '''
    dict_neutral_formula = parse_chemformula_dict(neutral_formula)
    adducts2get = []
    for x in __get_adduct_list__(mw, mode, primary_only):
        # ['mz', 'ion', 'delta_formula']
        result = add_formula_dict(dict_neutral_formula, x.delta_formula)
        if result:
            adducts2get.append( [x.mz, x.ion, dict_to_hill_formula(result)] )

    return adducts2get
